# -*- coding: utf-8 -*-

from datajuggler.core import aDict, uDict, iList
from datajuggler.keys import Keylist, Keypath

from datajuggler.utils import (
    StrCase, urange, omit_values, replace_values,
    split_chunks, add_df, df_compare, rename_duplicates,
)
from datajuggler.strings import (
    searchstr, substr, is_match_string, remove_accents,
)

__all__ = [
    "aDict",
    "uDict",
    "iList",
    "Keylist",
    "Keypath",
    "StrCase",
    "urange",
    "omit_values",
    "replace_values",
    "add_df",
    "df_compare",
    "split_chunks",
    "rename_duplicates",
    "searchstr",
    "substr",
    "is_match_string",
    "remove_accents",
]
